#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud GenAI Generative Model hook."""

from __future__ import annotations

import enum
import time
from typing import TYPE_CHECKING, Any

from google import genai

from airflow.providers.google.common.hooks.base_google import (
    PROVIDE_PROJECT_ID,
    GoogleBaseAsyncHook,
    GoogleBaseHook,
)

if TYPE_CHECKING:
    from google.genai.pagers import Pager
    from google.genai.types import (
        BatchJob,
        ContentListUnion,
        ContentListUnionDict,
        CountTokensConfigOrDict,
        CountTokensResponse,
        CreateBatchJobConfig,
        CreateCachedContentConfigOrDict,
        CreateTuningJobConfigOrDict,
        DeleteFileResponse,
        DeleteResourceJob,
        EmbedContentConfigOrDict,
        EmbedContentResponse,
        File,
        GenerateContentConfig,
        ListBatchJobsConfig,
        TuningDatasetOrDict,
        TuningJob,
    )


class BatchJobStatus(enum.Enum):
    """Possible states of batch job in Gemini Batch API."""

    SUCCEEDED = "JOB_STATE_SUCCEEDED"
    PENDING = "JOB_STATE_PENDING"
    FAILED = "JOB_STATE_FAILED"
    RUNNING = "JOB_STATE_RUNNING"
    CANCELLED = "JOB_STATE_CANCELLED"
    EXPIRED = "JOB_STATE_EXPIRED"


class GenAIGenerativeModelHook(GoogleBaseHook):
    """Class for Google Cloud Generative AI Vertex AI hook."""

    def get_genai_client(self, project_id: str, location: str):
        return genai.Client(
            vertexai=True,
            project=project_id,
            location=location,
        )

    @GoogleBaseHook.fallback_to_default_project_id
    def embed_content(
        self,
        model: str,
        location: str,
        contents: ContentListUnion | ContentListUnionDict | list[str],
        config: EmbedContentConfigOrDict | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
    ) -> EmbedContentResponse:
        """
        Generate embeddings for words, phrases, sentences, and code.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param model: Required. The model to use.
        :param contents: Optional. The contents to use for embedding.
        :param config: Optional. Configuration for embeddings.
        """
        client = self.get_genai_client(project_id=project_id, location=location)

        resp = client.models.embed_content(model=model, contents=contents, config=config)
        return resp

    @GoogleBaseHook.fallback_to_default_project_id
    def generate_content(
        self,
        location: str,
        model: str,
        contents: ContentListUnionDict,
        generation_config: GenerateContentConfig | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
    ) -> str:
        """
        Make an API request to generate content using a model.

        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param model: Required. The model to use.
        :param contents: Required. The multi-part content of a message that a user or a program
            gives to the generative model, in order to elicit a specific response.
        :param generation_config: Optional. Generation configuration settings.
        """
        client = self.get_genai_client(project_id=project_id, location=location)
        response = client.models.generate_content(
            model=model,
            contents=contents,
            config=generation_config,
        )

        return response.text

    @GoogleBaseHook.fallback_to_default_project_id
    def supervised_fine_tuning_train(
        self,
        source_model: str,
        location: str,
        training_dataset: TuningDatasetOrDict,
        tuning_job_config: CreateTuningJobConfigOrDict | dict[str, Any] | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
    ) -> TuningJob:
        """
        Create a tuning job to adapt model behavior with a labeled dataset.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param source_model: Required. A pre-trained model optimized for performing natural
            language tasks such as classification, summarization, extraction, content
            creation, and ideation.
        :param train_dataset: Required. Cloud Storage URI of your training dataset. The dataset
            must be formatted as a JSONL file. For best results, provide at least 100 to 500 examples.
        :param tuning_job_config: Optional. Configuration of the Tuning job to be created.
        """
        client = self.get_genai_client(project_id=project_id, location=location)

        tuning_job = client.tunings.tune(
            base_model=source_model,
            training_dataset=training_dataset,
            config=tuning_job_config,
        )

        # Poll until completion
        running = {"JOB_STATE_PENDING", "JOB_STATE_RUNNING"}
        while tuning_job.state in running:
            time.sleep(60)
            tuning_job = client.tunings.get(name=tuning_job.name)

        return tuning_job

    @GoogleBaseHook.fallback_to_default_project_id
    def count_tokens(
        self,
        location: str,
        model: str,
        contents: ContentListUnion | ContentListUnionDict,
        config: CountTokensConfigOrDict | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
    ) -> CountTokensResponse:
        """
        Use Count Tokens API to calculate the number of input tokens before sending a request to Gemini API.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param contents: Required. The multi-part content of a message that a user or a program
            gives to the generative model, in order to elicit a specific response.
        :param model: Required. Model,
            supporting prompts with text-only input, including natural language
            tasks, multi-turn text and code chat, and code generation. It can
            output text and code.
        :param config: Optional. Configuration for Count Tokens.
        """
        client = self.get_genai_client(project_id=project_id, location=location)
        response = client.models.count_tokens(
            model=model,
            contents=contents,
            config=config,
        )

        return response

    @GoogleBaseHook.fallback_to_default_project_id
    def create_cached_content(
        self,
        model: str,
        location: str,
        cached_content_config: CreateCachedContentConfigOrDict | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
    ) -> str:
        """
        Create CachedContent to reduce the cost of requests containing repeat content.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param model: Required. The name of the publisher model to use for cached content.
        :param cached_content_config: Optional. Configuration of the Cached Content.
        """
        client = self.get_genai_client(project_id=project_id, location=location)
        resp = client.caches.create(
            model=model,
            config=cached_content_config,
        )

        return resp.name


class GenAIGeminiAPIHook(GoogleBaseHook):
    """Class for Google Cloud Generative AI Gemini Developer API hook."""

    def __init__(self, gemini_api_key: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.gemini_api_key = gemini_api_key

    def get_genai_client(self):
        return genai.Client(
            api_key=self.gemini_api_key,
            vertexai=False,
        )

    def get_batch_job(
        self,
        job_name: str,
    ) -> BatchJob:
        """
        Get batch job using Gemini Batch API.

        :param job_name: Required. Batch job name.
        """
        client = self.get_genai_client()
        resp = client.batches.get(name=job_name)
        return resp

    def list_batch_jobs(
        self,
        list_batch_jobs_config: ListBatchJobsConfig | dict | None = None,
    ) -> Pager[BatchJob]:
        """
        Get list of batch jobs using Gemini Batch API.

        :param list_batch_jobs_config: Optional. Configuration of returned iterator.
        """
        client = self.get_genai_client()
        resp = client.batches.list(
            config=list_batch_jobs_config,
        )
        return resp

    def create_batch_job(
        self,
        model: str,
        source: list | str,
        create_batch_job_config: CreateBatchJobConfig | dict | None = None,
    ) -> BatchJob:
        """
        Create batch job using Gemini Batch API to process large-scale, non-urgent tasks.

        :param model: Required. Gemini model name to process requests.
        :param source: Required. Requests that will be sent to chosen model.
            Can be in format of Inline requests or file name.
        :param create_batch_job_config: Optional. Configuration parameters for batch job.
        """
        client = self.get_genai_client()
        resp = client.batches.create(
            model=model,
            src=source,
            config=create_batch_job_config,
        )
        return resp

    def delete_batch_job(
        self,
        job_name: str,
    ) -> DeleteResourceJob:
        """
        Delete batch job using Gemini Batch API.

        :param job_name: Required. Batch job name.
        """
        client = self.get_genai_client()
        resp = client.batches.delete(name=job_name)
        return resp

    def cancel_batch_job(
        self,
        job_name: str,
    ) -> None:
        """
        Cancel batch job using Gemini Batch API.

        :param job_name: Required. Batch job name.
        """
        client = self.get_genai_client()
        client.batches.cancel(
            name=job_name,
        )

    def create_embeddings(
        self,
        model: str,
        source: dict | str,
        create_embeddings_config: CreateBatchJobConfig | dict | None = None,
    ) -> BatchJob:
        """
        Create batch job for embeddings using Gemini Batch API to process large-scale, non-urgent tasks.

        :param model: Required. Gemini model name to process requests.
        :param source: Required. Requests that will be sent to chosen model.
            Can be in format of Inline requests or file name.
        :param create_embeddings_config: Optional. Configuration parameters for embeddings batch job.
        """
        client = self.get_genai_client()
        input_type = "inlined_requests"

        if isinstance(source, str):
            input_type = "file_name"

        self.log.info("Using %s to create embeddings", input_type)

        resp = client.batches.create_embeddings(
            model=model,
            src={input_type: source},
            config=create_embeddings_config,
        )
        return resp

    def upload_file(self, path_to_file: str, upload_file_config: dict | None = None) -> File:
        """
        Upload file for batch job or embeddings batch job using Gemini Files API.

        :param path_to_file: Required. Path to file on local filesystem.
        :param upload_file_config: Optional. Configuration for file upload.
        """
        client = self.get_genai_client()

        if upload_file_config is None:
            self.log.info("Default configuration will be used to upload file")
            try:
                file_name, file_type = path_to_file.split("/")[-1].split(".")
                upload_file_config = {"display_name": file_name, "mime_type": file_type}
            except ValueError as exc:
                raise ValueError(
                    "Error during unpacking file name or mime type. Please check file path"
                ) from exc

        resp = client.files.upload(
            file=path_to_file,
            config=upload_file_config,
        )
        return resp

    def get_file(self, file_name: str) -> File:
        """
        Get file's metadata for batch job or embeddings batch job using Gemini Files API.

        :param file_name: Required. Name of the file in Gemini Files API.
        """
        client = self.get_genai_client()
        resp = client.files.get(name=file_name)
        return resp

    def download_file(self, file_name: str) -> bytes:
        """
        Download file for batch job or embeddings batch job using Gemini Files API.

        :param file_name: Required. Name of the file in Gemini Files API.
        """
        client = self.get_genai_client()
        resp = client.files.download(file=file_name)
        return resp

    def list_files(self) -> Pager[File]:
        """List files for stored in Gemini Files API."""
        client = self.get_genai_client()
        resp = client.files.list()
        return resp

    def delete_file(self, file_name: str) -> DeleteFileResponse:
        """
        Delete file from Gemini Files API storage.

        :param file_name: Required. Name of the file in Gemini Files API.
        """
        client = self.get_genai_client()
        resp = client.files.delete(name=file_name)
        return resp


class GenAIGeminiAPIAsyncHook(GoogleBaseAsyncHook):
    """Class for Google Cloud Generative AI Gemini Developer Async API hook."""

    sync_hook_class = GenAIGeminiAPIHook

    def __init__(self, gemini_api_key: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.gemini_api_key = gemini_api_key

    async def get_async_client(self):
        return genai.Client(
            api_key=self.gemini_api_key,
            vertexai=False,
        ).aio

    async def create_batch_job(
        self,
        model: str,
        source: list | str,
        create_batch_job_config: CreateBatchJobConfig | dict | None = None,
    ) -> BatchJob:
        """
        Create batch job asynchronously using Gemini Batch API to process large-scale, non-urgent tasks.

        :param model: Required. Gemini model name to process requests.
        :param source: Required. Requests that will be sent to chosen model.
            Can be in format of Inline requests or file name.
        :param create_batch_job_config: Optional. Configuration parameters for batch job.
        """
        async_client = await self.get_async_client()
        async_job = await async_client.batches.create(model=model, src=source, config=create_batch_job_config)
        return async_job

    async def get_batch_job(
        self,
        job_name: str,
    ) -> BatchJob:
        """
        Get batch job using Gemini Batch API asynchronously.

        :param job_name: Required. Batch job name.
        """
        async_client = await self.get_async_client()
        resp = await async_client.batches.get(name=job_name)
        return resp

    async def create_embeddings_batch_job(
        self,
        model: str,
        source: dict | str,
        create_embeddings_config: CreateBatchJobConfig | dict | None = None,
    ) -> BatchJob:
        """
        Create batch job for embeddings asynchronously using Gemini Batch API to process large-scale, non-urgent tasks.

        :param model: Required. Gemini model name to process requests.
        :param source: Required. Requests that will be sent to chosen model.
            Can be in format of Inline requests or file name.
        :param create_embeddings_config: Optional. Configuration parameters for embeddings batch job.
        """
        async_client = await self.get_async_client()
        input_type = "inlined_requests"

        if isinstance(source, str):
            input_type = "file_name"

        self.log.info("Using %s to create embeddings", input_type)

        resp = await async_client.batches.create_embeddings(
            model=model,
            src={input_type: source},
            config=create_embeddings_config,
        )
        return resp



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import warnings
from unittest import mock

import pytest

from airflow.providers.google.cloud.transfers.calendar_to_gcs import GoogleCalendarToGCSOperator

API_VERSION = "v3"
CALENDAR_ID = "calendar_id"
EVENT = {"id": "123"}
BUCKET = "destination_bucket"
PATH = "path/to/reports"
GCP_CONN_ID = "test"
IMPERSONATION_CHAIN = ["account1", "account2"]


class TestGoogleCalendarToGCSOperator:
    @mock.patch("airflow.providers.google.cloud.transfers.calendar_to_gcs.NamedTemporaryFile")
    @mock.patch("airflow.providers.google.cloud.transfers.calendar_to_gcs.json")
    @mock.patch("airflow.providers.google.cloud.transfers.calendar_to_gcs.GCSHook")
    @mock.patch("airflow.providers.google.cloud.transfers.calendar_to_gcs.GoogleCalendarHook")
    def test_upload_data(self, mock_calendar_hook, mock_gcs_hook, mock_json, mock_temp_file):
        expected_dest_file = f"{CALENDAR_ID}.json".replace(" ", "_")
        expected_dest_file = f"{PATH.strip('/')}/{expected_dest_file}"
        expected_gcs_uri = f"gs://{BUCKET}/{expected_dest_file}"

        mock_calendar_hook.return_value.get_events.return_value = [EVENT]
        mock_gcs_hook.return_value.upload.return_value = None
        file_handle = mock_temp_file.return_value.__enter__.return_value
        file_handle.name = "file_name"

        op = GoogleCalendarToGCSOperator(
            api_version=API_VERSION,
            task_id="test_task",
            calendar_id=CALENDAR_ID,
            destination_bucket=BUCKET,
            destination_path=PATH,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        result = op._upload_data(
            events=[EVENT],
        )

        # Test writing to file
        file_handle.flush.assert_called_once_with()

        # Test GCS Hook
        mock_gcs_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        # Assert tuple of (dest_file_name, gcs_uri) is returned
        assert result == (expected_dest_file, expected_gcs_uri)

    @mock.patch("airflow.providers.google.cloud.transfers.calendar_to_gcs.GoogleCalendarHook")
    @mock.patch(
        "airflow.providers.google.cloud.transfers.calendar_to_gcs.GoogleCalendarToGCSOperator._upload_data"
    )
    def test_execute_with_unwrap_single_true(self, mock_upload_data, mock_calendar_hook):
        context = {}
        data = [EVENT]
        expected_dest_file_name = f"{PATH.strip('/')}/{CALENDAR_ID}.json"
        mock_upload_data.return_value = (expected_dest_file_name, f"gs://{BUCKET}/{PATH}")
        mock_calendar_hook.return_value.get_events.return_value = data

        op = GoogleCalendarToGCSOperator(
            api_version=API_VERSION,
            task_id="test_task",
            calendar_id=CALENDAR_ID,
            destination_bucket=BUCKET,
            destination_path=PATH,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            unwrap_single=True,
        )
        result = op.execute(context)

        mock_calendar_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_calendar_hook.return_value.get_events.assert_called_once_with(
            calendar_id=CALENDAR_ID,
            i_cal_uid=None,
            max_attendees=None,
            max_results=None,
            order_by=None,
            private_extended_property=None,
            q=None,
            shared_extended_property=None,
            show_deleted=None,
            show_hidden_invitation=None,
            single_events=None,
            sync_token=None,
            time_max=None,
            time_min=None,
            time_zone=None,
            updated_min=None,
        )

        mock_upload_data.assert_called_once_with(data)
        # Assert dest_file_name is returned when unwrap_single=True and return_gcs_uri=False (default)
        assert result == expected_dest_file_name

    @mock.patch("airflow.providers.google.cloud.transfers.calendar_to_gcs.GoogleCalendarHook")
    @mock.patch(
        "airflow.providers.google.cloud.transfers.calendar_to_gcs.GoogleCalendarToGCSOperator._upload_data"
    )
    def test_execute_with_unwrap_single_false(self, mock_upload_data, mock_calendar_hook):
        context = {}
        data = [EVENT]
        expected_dest_file_name = f"{PATH.strip('/')}/{CALENDAR_ID}.json"
        expected_gcs_uri = f"gs://{BUCKET}/{expected_dest_file_name}"
        mock_upload_data.return_value = (expected_dest_file_name, expected_gcs_uri)
        mock_calendar_hook.return_value.get_events.return_value = data

        op = GoogleCalendarToGCSOperator(
            api_version=API_VERSION,
            task_id="test_task",
            calendar_id=CALENDAR_ID,
            destination_bucket=BUCKET,
            destination_path=PATH,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            unwrap_single=False,
            return_gcs_uri=True,
        )
        result = op.execute(context)

        mock_calendar_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_calendar_hook.return_value.get_events.assert_called_once_with(
            calendar_id=CALENDAR_ID,
            i_cal_uid=None,
            max_attendees=None,
            max_results=None,
            order_by=None,
            private_extended_property=None,
            q=None,
            shared_extended_property=None,
            show_deleted=None,
            show_hidden_invitation=None,
            single_events=None,
            sync_token=None,
            time_max=None,
            time_min=None,
            time_zone=None,
            updated_min=None,
        )

        mock_upload_data.assert_called_once_with(data)
        # Assert list of GCS URIs is returned when unwrap_single=False and return_gcs_uri=True
        assert result == [expected_gcs_uri]

    @mock.patch("airflow.providers.google.cloud.transfers.calendar_to_gcs.GoogleCalendarHook")
    @mock.patch(
        "airflow.providers.google.cloud.transfers.calendar_to_gcs.GoogleCalendarToGCSOperator._upload_data"
    )
    def test_execute_with_return_gcs_uri_true_and_unwrap_single_false(
        self, mock_upload_data, mock_calendar_hook
    ):
        context = {}
        data = [EVENT]
        expected_dest_file_name = f"{PATH.strip('/')}/{CALENDAR_ID}.json"
        expected_gcs_uri = f"gs://{BUCKET}/{expected_dest_file_name}"
        mock_upload_data.return_value = (expected_dest_file_name, expected_gcs_uri)
        mock_calendar_hook.return_value.get_events.return_value = data

        op = GoogleCalendarToGCSOperator(
            api_version=API_VERSION,
            task_id="test_task",
            calendar_id=CALENDAR_ID,
            destination_bucket=BUCKET,
            destination_path=PATH,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            unwrap_single=False,
            return_gcs_uri=True,
        )
        result = op.execute(context)

        mock_calendar_hook.assert_called_once_with(
            api_version=API_VERSION,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_calendar_hook.return_value.get_events.assert_called_once_with(
            calendar_id=CALENDAR_ID,
            i_cal_uid=None,
            max_attendees=None,
            max_results=None,
            order_by=None,
            private_extended_property=None,
            q=None,
            shared_extended_property=None,
            show_deleted=None,
            show_hidden_invitation=None,
            single_events=None,
            sync_token=None,
            time_max=None,
            time_min=None,
            time_zone=None,
            updated_min=None,
        )

        mock_upload_data.assert_called_once_with(data)
        # Assert list of GCS URIs is returned when return_gcs_uri=True
        assert result == [expected_gcs_uri]

    def test_execute_with_return_gcs_uri_false_and_unwrap_single_false_raises_error(self):
        """Test that return_gcs_uri=False together with unwrap_single=False raises ValueError."""
        with pytest.raises(
            ValueError, match="return_gcs_uri cannot be False together with unwrap_single=False"
        ):
            GoogleCalendarToGCSOperator(
                api_version=API_VERSION,
                task_id="test_task",
                calendar_id=CALENDAR_ID,
                destination_bucket=BUCKET,
                destination_path=PATH,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
                unwrap_single=False,
                return_gcs_uri=False,
            )

    def test_unwrap_single_none_emits_warning(self):
        """Test that unwrap_single=None emits a deprecation warning."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            GoogleCalendarToGCSOperator(
                api_version=API_VERSION,
                task_id="test_task",
                calendar_id=CALENDAR_ID,
                destination_bucket=BUCKET,
                destination_path=PATH,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
                unwrap_single=None,
                return_gcs_uri=True,
            )
            assert len(w) == 1
            assert issubclass(w[0].category, FutureWarning)
            assert "unwrap_single will change from True to False" in str(w[0].message)

    def test_return_gcs_uri_none_emits_warning(self):
        """Test that return_gcs_uri=None emits a deprecation warning."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            GoogleCalendarToGCSOperator(
                api_version=API_VERSION,
                task_id="test_task",
                calendar_id=CALENDAR_ID,
                destination_bucket=BUCKET,
                destination_path=PATH,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
                unwrap_single=True,
                return_gcs_uri=None,
            )
            assert len(w) == 1
            assert issubclass(w[0].category, FutureWarning)
            assert "return_gcs_uri parameter is deprecated" in str(w[0].message)

    def test_both_none_emits_both_warnings(self):
        """Test that both None values emit both deprecation warnings."""
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            # Create operator but don't need to use it - just checking for warnings
            GoogleCalendarToGCSOperator(
                api_version=API_VERSION,
                task_id="test_task",
                calendar_id=CALENDAR_ID,
                destination_bucket=BUCKET,
                destination_path=PATH,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
                unwrap_single=None,
                return_gcs_uri=None,
            )
            # Should have 2 warnings and not raise ValueError since return_gcs_uri gets set to False
            # when None, and unwrap_single gets set to True when None
            assert len(w) == 2
            assert all(issubclass(warning.category, FutureWarning) for warning in w)

    def test_return_gcs_uri_true_with_unwrap_single_true_is_valid(self):
        """Test that return_gcs_uri=True with unwrap_single=True is valid (opposite of old logic)."""
        # This should NOT raise an error (the old test was checking the wrong condition)
        op = GoogleCalendarToGCSOperator(
            api_version=API_VERSION,
            task_id="test_task",
            calendar_id=CALENDAR_ID,
            destination_bucket=BUCKET,
            destination_path=PATH,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            unwrap_single=True,
            return_gcs_uri=True,
        )
        assert op.unwrap_single is True
        assert op.return_gcs_uri is True



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from google.genai.errors import ClientError
from google.genai.types import (
    Content,
    CreateCachedContentConfig,
    GenerateContentConfig,
    GoogleSearch,
    Part,
    Tool,
    TuningDataset,
)

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.operators.gen_ai import (
    GenAICountTokensOperator,
    GenAICreateCachedContentOperator,
    GenAIGeminiCancelBatchJobOperator,
    GenAIGeminiCreateBatchJobOperator,
    GenAIGeminiCreateEmbeddingsBatchJobOperator,
    GenAIGeminiDeleteBatchJobOperator,
    GenAIGeminiDeleteFileOperator,
    GenAIGeminiGetBatchJobOperator,
    GenAIGeminiGetFileOperator,
    GenAIGeminiListBatchJobsOperator,
    GenAIGeminiListFilesOperator,
    GenAIGeminiUploadFileOperator,
    GenAIGenerateContentOperator,
    GenAIGenerateEmbeddingsOperator,
    GenAISupervisedFineTuningTrainOperator,
)

GEN_AI_PATH = "airflow.providers.google.cloud.operators.gen_ai.{}"

TASK_ID = "test_task_id"
GCP_PROJECT = "test-project"
GCP_LOCATION = "test-location"
GCP_CONN_ID = "test-conn"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
CACHED_SYSTEM_INSTRUCTION = """
You are an expert researcher. You always stick to the facts in the sources provided, and never make up new facts.
Now look at these research papers, and answer the following questions.
"""
CACHED_CONTENT_CONFIG = CreateCachedContentConfig(
    contents=[
        Content(
            role="user",
            parts=[
                Part.from_uri(
                    file_uri="gs://cloud-samples-data/generative-ai/pdf/2312.11805v3.pdf",
                    mime_type="application/pdf",
                ),
                Part.from_uri(
                    file_uri="gs://cloud-samples-data/generative-ai/pdf/2403.05530.pdf",
                    mime_type="application/pdf",
                ),
            ],
        )
    ],
    system_instruction=CACHED_SYSTEM_INSTRUCTION,
    display_name="test-cache",
    ttl="3600s",
)
EMBEDDING_MODEL = "textembedding-gecko"
GEMINI_MODEL = "gemini-pro"
CONTENTS = ["In 10 words or less, what is Apache Airflow?"]
CONTENT_GENERATION_CONFIG = GenerateContentConfig(
    max_output_tokens=256,
    top_p=0.95,
    temperature=0.0,
    tools=[Tool(google_search=GoogleSearch())],
)
TUNING_JOB_CONFIG = TuningDataset(
    gcs_uri="gs://cloud-samples-data/ai-platform/generative_ai/gemini-1_5/text/sft_train_data.jsonl",
)
TUNING_TRAINING_DATASET = "gs://cloud-samples-data/ai-platform/generative_ai/sft_train_data.jsonl"
GENERATE_FROM_CACHED_MODEL_CONFIG = {
    "cached_content": "cached_name",
}

TEST_BATCH_JOB_INLINED_REQUESTS = [
    {"contents": [{"parts": [{"text": "Tell me a one-sentence joke."}], "role": "user"}]},
    {"contents": [{"parts": [{"text": "Why is the sky blue?"}], "role": "user"}]},
]

TEST_EMBEDDINGS_JOB_INLINED_REQUESTS = {
    "contents": [{"parts": [{"text": "Why is the sky blue?"}], "role": "user"}]
}
TEST_GEMINI_API_KEY = "test-key"
TEST_GEMINI_MODEL = "test-gemini-model"
TEST_BATCH_JOB_NAME = "test-name"
TEST_FILE_NAME = "test-file"
TEST_FILE_PATH = "test/path/to/file"
TEST_BATCH_JOB_RESPONSE = {
    "src": None,
    "dest": "test-batch-job-destination",
    "name": "test-name",
    "error": None,
    "model": "test-model",
    "state": "JOB_STATE_SUCCEEDED",
    "end_time": "test-end-datetime",
    "start_time": None,
    "create_time": "test-create-datetime",
    "update_time": "test-update-datetime",
    "display_name": "test-display-name",
    "completion_stats": None,
}


def assert_warning(msg: str, warnings):
    assert any(msg in str(w) for w in warnings)


class TestGenAIGenerateEmbeddingsOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGenerativeModelHook"))
    def test_execute(self, mock_hook):
        op = GenAIGenerateEmbeddingsOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=CONTENTS,
            model=EMBEDDING_MODEL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.embed_content.assert_called_once_with(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=CONTENTS,
            model=EMBEDDING_MODEL,
            config=None,
        )


class TestGenAIGenerateContentOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGenerativeModelHook"))
    def test_execute(self, mock_hook):
        op = GenAIGenerateContentOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=CONTENTS,
            generation_config=CONTENT_GENERATION_CONFIG,
            model=GEMINI_MODEL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.generate_content.assert_called_once_with(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=CONTENTS,
            generation_config=CONTENT_GENERATION_CONFIG,
            model=GEMINI_MODEL,
        )


class TestGenAISupervisedFineTuningTrainOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGenerativeModelHook"))
    def test_execute(
        self,
        mock_hook,
    ):
        op = GenAISupervisedFineTuningTrainOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            source_model=GEMINI_MODEL,
            training_dataset=TUNING_TRAINING_DATASET,
            tuning_job_config=TUNING_JOB_CONFIG,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.supervised_fine_tuning_train.assert_called_once_with(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            source_model=GEMINI_MODEL,
            training_dataset=TUNING_TRAINING_DATASET,
            tuning_job_config=TUNING_JOB_CONFIG,
        )


class TestGenAICountTokensOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGenerativeModelHook"))
    def test_execute(self, mock_hook):
        op = GenAICountTokensOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=CONTENTS,
            model=GEMINI_MODEL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.count_tokens.assert_called_once_with(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=CONTENTS,
            model=GEMINI_MODEL,
            config=None,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGenerativeModelHook"))
    def test_execute_propagates_client_error(self, mock_hook):
        """Test that GenAICountTokensOperator propagates ClientError from the hook."""
        # Bypass __init__ to avoid version-dependent constructor signature differences
        # between google-genai versions (e.g., 1.2.0 vs 1.63.0).
        mock_hook.return_value.count_tokens.side_effect = ClientError.__new__(ClientError)

        op = GenAICountTokensOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            contents=CONTENTS,
            model=GEMINI_MODEL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        with pytest.raises(ClientError):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )


class TestGenAICreateCachedContentOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGenerativeModelHook"))
    def test_execute(self, mock_hook):
        op = GenAICreateCachedContentOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=GEMINI_MODEL,
            cached_content_config=CACHED_CONTENT_CONFIG,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.create_cached_content.assert_called_once_with(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=GEMINI_MODEL,
            cached_content_config=CACHED_CONTENT_CONFIG,
        )


class TestGenAIGenerateFromCachedContentOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGenerativeModelHook"))
    def test_execute(self, mock_hook):
        op = GenAIGenerateContentOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=GEMINI_MODEL,
            contents=CONTENTS,
            generation_config=GENERATE_FROM_CACHED_MODEL_CONFIG,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.return_value.generate_content.assert_called_once_with(
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=GEMINI_MODEL,
            contents=CONTENTS,
            generation_config=GENERATE_FROM_CACHED_MODEL_CONFIG,
        )


class TestGenAIGeminiCreateBatchJobOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute(self, mock_hook):
        op = GenAIGeminiCreateBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=TEST_GEMINI_MODEL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            input_source=TEST_BATCH_JOB_INLINED_REQUESTS,
            gemini_api_key=TEST_GEMINI_API_KEY,
            wait_until_complete=False,
            retrieve_result=False,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        mock_hook.return_value.create_batch_job.assert_called_once_with(
            source=TEST_BATCH_JOB_INLINED_REQUESTS,
            model=TEST_GEMINI_MODEL,
            create_batch_job_config=None,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_return_value(self, mock_hook):
        expected_return = TEST_BATCH_JOB_RESPONSE

        mock_job = mock.MagicMock()
        mock_job.model_dump.return_value = expected_return
        mock_hook.return_value.create_batch_job.return_value = mock_job

        op = GenAIGeminiCreateBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=TEST_GEMINI_MODEL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            input_source=TEST_BATCH_JOB_INLINED_REQUESTS,
            gemini_api_key=TEST_GEMINI_API_KEY,
            deferrable=False,
            wait_until_complete=False,
        )

        result = op.execute(context={"ti": mock.MagicMock()})

        assert result == expected_return
        mock_job.model_dump.assert_called_once_with(mode="json")

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_complete_return_value(self, mock_hook):
        expected_return = TEST_BATCH_JOB_RESPONSE

        event = {"status": "success", "job_name": "test-name"}

        mock_job = mock.MagicMock()
        mock_job.model_dump.return_value = expected_return
        mock_hook.return_value.get_batch_job.return_value = mock_job

        op = GenAIGeminiCreateBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=TEST_GEMINI_MODEL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            input_source=TEST_BATCH_JOB_INLINED_REQUESTS,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        result = op.execute_complete(context={"ti": mock.MagicMock()}, event=event)

        assert result == expected_return
        mock_hook.return_value.get_batch_job.assert_called_once_with("test-name")
        mock_job.model_dump.assert_called_once_with(mode="json")

    def test_init_retrieve_result_and_not_wait_until_complete_raises_airflow_exception(self):
        with pytest.raises(AirflowException):
            GenAIGeminiCreateBatchJobOperator(
                task_id=TASK_ID,
                project_id=GCP_PROJECT,
                location=GCP_LOCATION,
                model=TEST_GEMINI_MODEL,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
                input_source=TEST_BATCH_JOB_INLINED_REQUESTS,
                gemini_api_key=TEST_GEMINI_API_KEY,
                wait_until_complete=False,
                retrieve_result=True,
            )

    def test_init_input_source_not_string_raises_airflow_exception(self):
        with pytest.raises(AirflowException):
            GenAIGeminiCreateBatchJobOperator(
                task_id=TASK_ID,
                project_id=GCP_PROJECT,
                location=GCP_LOCATION,
                model=TEST_GEMINI_MODEL,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
                input_source=TEST_BATCH_JOB_INLINED_REQUESTS,
                gemini_api_key=TEST_GEMINI_API_KEY,
                wait_until_complete=False,
                results_folder=TEST_FILE_PATH,
            )

    def test_init_results_folder_not_exists_raises_airflow_exception(self):
        with pytest.raises(AirflowException):
            GenAIGeminiCreateBatchJobOperator(
                task_id=TASK_ID,
                project_id=GCP_PROJECT,
                location=GCP_LOCATION,
                model=TEST_GEMINI_MODEL,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
                input_source=TEST_FILE_NAME,
                gemini_api_key=TEST_GEMINI_API_KEY,
                wait_until_complete=False,
                results_folder=TEST_FILE_PATH,
            )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test__wait_until_complete_exception_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiCreateBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=TEST_GEMINI_MODEL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            input_source=TEST_BATCH_JOB_INLINED_REQUESTS,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.get_batch_job.side_effect = Exception()

        with pytest.raises(AirflowException):
            op._wait_until_complete(job=mock.MagicMock())

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_exception_error_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiCreateBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=TEST_GEMINI_MODEL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            input_source=TEST_BATCH_JOB_INLINED_REQUESTS,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.create_batch_job.side_effect = Exception()

        with pytest.raises(AirflowException):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.create_batch_job.assert_called_once_with(
            source=TEST_BATCH_JOB_INLINED_REQUESTS,
            model=TEST_GEMINI_MODEL,
            create_batch_job_config=None,
        )

    def test_execute_complete_error_status_raises_airflow_exception(self):
        op = GenAIGeminiCreateBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            model=TEST_GEMINI_MODEL,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            input_source=TEST_BATCH_JOB_INLINED_REQUESTS,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        event = {"status": "error", "message": "test-message"}

        with pytest.raises(AirflowException):
            op.execute_complete(context={"ti": mock.MagicMock()}, event=event)


class TestGenAIGeminiGetBatchJobOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute(self, mock_hook):
        op = GenAIGeminiGetBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            job_name=TEST_BATCH_JOB_NAME,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        mock_hook.return_value.get_batch_job.assert_called_once_with(
            job_name=TEST_BATCH_JOB_NAME,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_return_value(self, mock_hook):
        expected_return = TEST_BATCH_JOB_RESPONSE

        mock_job = mock.MagicMock()
        mock_job.model_dump.return_value = expected_return
        mock_hook.return_value.get_batch_job.return_value = mock_job

        op = GenAIGeminiGetBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            job_name=TEST_BATCH_JOB_NAME,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        result = op.execute(context={"ti": mock.MagicMock()})

        assert result == expected_return
        mock_job.model_dump.assert_called_once_with(mode="json")

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_value_error_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiGetBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            job_name=TEST_BATCH_JOB_NAME,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.get_batch_job.side_effect = ValueError()

        with pytest.raises(AirflowException):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.get_batch_job.assert_called_once_with(
            job_name=TEST_BATCH_JOB_NAME,
        )


class TestGenAIGeminiListBatchJobsOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute(self, mock_hook):
        op = GenAIGeminiListBatchJobsOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        mock_hook.return_value.list_batch_jobs.assert_called_once_with(list_batch_jobs_config=None)


class TestGenAIGeminiDeleteBatchJobOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute(self, mock_hook):
        mock_hook.return_value.delete_batch_job.return_value = mock.MagicMock(error=False)
        op = GenAIGeminiDeleteBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            job_name=TEST_BATCH_JOB_NAME,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        mock_hook.return_value.delete_batch_job.assert_called_once_with(
            job_name=TEST_BATCH_JOB_NAME,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_value_error_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiDeleteBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            job_name=TEST_BATCH_JOB_NAME,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.delete_batch_job.side_effect = ValueError()

        with pytest.raises(AirflowException):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.delete_batch_job.assert_called_once_with(
            job_name=TEST_BATCH_JOB_NAME,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_job_error_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiDeleteBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            job_name=TEST_BATCH_JOB_NAME,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.delete_batch_job.return_value = mock.MagicMock(error="Test error")

        with pytest.raises(AirflowException):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.delete_batch_job.assert_called_once_with(
            job_name=TEST_BATCH_JOB_NAME,
        )


class TestGenAIGeminiCancelBatchJobOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute(self, mock_hook):
        op = GenAIGeminiCancelBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            job_name=TEST_BATCH_JOB_NAME,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        mock_hook.return_value.cancel_batch_job.assert_called_once_with(
            job_name=TEST_BATCH_JOB_NAME,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_value_error_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiCancelBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            job_name=TEST_BATCH_JOB_NAME,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.cancel_batch_job.side_effect = ValueError()

        with pytest.raises(AirflowException):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.cancel_batch_job.assert_called_once_with(
            job_name=TEST_BATCH_JOB_NAME,
        )


class TestGenAIGeminiCreateEmbeddingsBatchJobOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute(self, mock_hook):
        op = GenAIGeminiCreateEmbeddingsBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            input_source=TEST_EMBEDDINGS_JOB_INLINED_REQUESTS,
            model=EMBEDDING_MODEL,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            wait_until_complete=False,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        mock_hook.return_value.create_embeddings.assert_called_once_with(
            source=TEST_EMBEDDINGS_JOB_INLINED_REQUESTS,
            model=EMBEDDING_MODEL,
            create_embeddings_config=None,
        )

    def test_init_retrieve_result_and_not_wait_until_complete_raises_airflow_exception(self):
        with pytest.raises(AirflowException):
            GenAIGeminiCreateEmbeddingsBatchJobOperator(
                task_id=TASK_ID,
                project_id=GCP_PROJECT,
                location=GCP_LOCATION,
                input_source=TEST_EMBEDDINGS_JOB_INLINED_REQUESTS,
                model=EMBEDDING_MODEL,
                gemini_api_key=TEST_GEMINI_API_KEY,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
                wait_until_complete=False,
                retrieve_result=True,
            )

    def test_init_input_source_not_string_raises_airflow_exception(self):
        with pytest.raises(AirflowException):
            GenAIGeminiCreateEmbeddingsBatchJobOperator(
                task_id=TASK_ID,
                project_id=GCP_PROJECT,
                location=GCP_LOCATION,
                input_source=TEST_EMBEDDINGS_JOB_INLINED_REQUESTS,
                model=EMBEDDING_MODEL,
                gemini_api_key=TEST_GEMINI_API_KEY,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
                wait_until_complete=False,
                results_folder=TEST_FILE_PATH,
            )

    def test_init_results_folder_not_exists_raises_airflow_exception(self):
        with pytest.raises(AirflowException):
            GenAIGeminiCreateEmbeddingsBatchJobOperator(
                task_id=TASK_ID,
                project_id=GCP_PROJECT,
                location=GCP_LOCATION,
                input_source=TEST_FILE_NAME,
                model=EMBEDDING_MODEL,
                gemini_api_key=TEST_GEMINI_API_KEY,
                gcp_conn_id=GCP_CONN_ID,
                impersonation_chain=IMPERSONATION_CHAIN,
                wait_until_complete=False,
                results_folder=TEST_FILE_PATH,
            )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test__wait_until_complete_exception_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiCreateEmbeddingsBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            input_source=TEST_EMBEDDINGS_JOB_INLINED_REQUESTS,
            model=EMBEDDING_MODEL,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_batch_job.side_effect = Exception()

        with pytest.raises(AirflowException):
            op._wait_until_complete(job=mock.MagicMock())

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_exception_error_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiCreateEmbeddingsBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            input_source=TEST_EMBEDDINGS_JOB_INLINED_REQUESTS,
            model=EMBEDDING_MODEL,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            wait_until_complete=False,
        )

        mock_hook.return_value.create_embeddings.side_effect = Exception()

        with pytest.raises(AirflowException):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.create_embeddings.assert_called_once_with(
            source=TEST_EMBEDDINGS_JOB_INLINED_REQUESTS,
            model=EMBEDDING_MODEL,
            create_embeddings_config=None,
        )

    def test_execute_complete_error_status_raises_airflow_exception(self):
        op = GenAIGeminiCreateEmbeddingsBatchJobOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            input_source=TEST_EMBEDDINGS_JOB_INLINED_REQUESTS,
            model=EMBEDDING_MODEL,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        event = {"status": "error", "message": "test-message"}

        with pytest.raises(AirflowException):
            op.execute_complete(context={"ti": mock.MagicMock()}, event=event)


class TestGenAIGeminiGetFileOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute(self, mock_hook):
        op = GenAIGeminiGetFileOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            gemini_api_key=TEST_GEMINI_API_KEY,
            file_name=TEST_FILE_NAME,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        mock_hook.return_value.get_file.assert_called_once_with(
            file_name=TEST_FILE_NAME,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_client_error_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiGetFileOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            gemini_api_key=TEST_GEMINI_API_KEY,
            file_name=TEST_FILE_NAME,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.get_file.side_effect = ClientError.__new__(ClientError)

        with pytest.raises(AirflowException):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.get_file.assert_called_once_with(
            file_name=TEST_FILE_NAME,
        )


class TestGenAIGeminiUploadFileOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute(self, mock_hook):
        op = GenAIGeminiUploadFileOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            file_path=TEST_FILE_PATH,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        mock_hook.return_value.upload_file.assert_called_once_with(
            path_to_file=TEST_FILE_PATH,
            upload_file_config=None,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_runtime_error_raises_runtime_error(self, mock_hook):
        op = GenAIGeminiUploadFileOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            file_path=TEST_FILE_PATH,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.upload_file.side_effect = RuntimeError()

        with pytest.raises(RuntimeError):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.upload_file.assert_called_once_with(
            path_to_file=TEST_FILE_PATH,
            upload_file_config=None,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_value_error_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiUploadFileOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            file_path=TEST_FILE_PATH,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.upload_file.side_effect = ValueError()

        with pytest.raises(AirflowException):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.upload_file.assert_called_once_with(
            path_to_file=TEST_FILE_PATH,
            upload_file_config=None,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_file_not_found_error_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiUploadFileOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            file_path=TEST_FILE_PATH,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.upload_file.side_effect = FileNotFoundError()

        with pytest.raises(AirflowException):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.upload_file.assert_called_once_with(
            path_to_file=TEST_FILE_PATH,
            upload_file_config=None,
        )


class TestGenAIGeminiListFilesOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute(self, mock_hook):
        op = GenAIGeminiListFilesOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        mock_hook.return_value.list_files.assert_called_once_with()


class TestGenAIGeminiDeleteFileOperator:
    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute(self, mock_hook):
        op = GenAIGeminiDeleteFileOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            file_name=TEST_FILE_NAME,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        op.execute(context={"ti": mock.MagicMock()})
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )
        mock_hook.return_value.delete_file.assert_called_once_with(
            file_name=TEST_FILE_NAME,
        )

    @mock.patch(GEN_AI_PATH.format("GenAIGeminiAPIHook"))
    def test_execute_client_error_raises_airflow_exception(self, mock_hook):
        op = GenAIGeminiDeleteFileOperator(
            task_id=TASK_ID,
            project_id=GCP_PROJECT,
            location=GCP_LOCATION,
            file_name=TEST_FILE_NAME,
            gemini_api_key=TEST_GEMINI_API_KEY,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.return_value.delete_file.side_effect = ClientError.__new__(ClientError)

        with pytest.raises(AirflowException):
            op.execute(context={"ti": mock.MagicMock()})

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            gemini_api_key=TEST_GEMINI_API_KEY,
        )

        mock_hook.return_value.delete_file.assert_called_once_with(
            file_name=TEST_FILE_NAME,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Vertex AI Generative Model hook."""

from __future__ import annotations

from typing import Any

import vertexai
from vertexai.generative_models import GenerativeModel
from vertexai.language_models import TextEmbeddingModel
from vertexai.preview import generative_models as preview_generative_model
from vertexai.preview.caching import CachedContent
from vertexai.preview.evaluation import EvalResult, EvalTask

from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID, GoogleBaseHook


class GenerativeModelHook(GoogleBaseHook):
    """Hook for Google Cloud Vertex AI Generative Model APIs."""

    def get_text_embedding_model(self, pretrained_model: str):
        """Return a Model Garden Model object based on Text Embedding."""
        model = TextEmbeddingModel.from_pretrained(pretrained_model)
        return model

    def get_generative_model(
        self,
        pretrained_model: str,
        system_instruction: Any | None = None,
        generation_config: dict | None = None,
        safety_settings: dict | None = None,
        tools: list | None = None,
    ) -> GenerativeModel:
        """Return a Generative Model object."""
        model = GenerativeModel(
            model_name=pretrained_model,
            system_instruction=system_instruction,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
        )
        return model

    def get_eval_task(
        self,
        dataset: dict,
        metrics: list,
        experiment: str,
    ) -> EvalTask:
        """Return an EvalTask object."""
        eval_task = EvalTask(
            dataset=dataset,
            metrics=metrics,
            experiment=experiment,
        )
        return eval_task

    def get_cached_context_model(
        self,
        cached_content_name: str,
    ) -> Any:
        """Return a Generative Model with Cached Context."""
        cached_content = CachedContent(cached_content_name=cached_content_name)

        cached_context_model = preview_generative_model.GenerativeModel.from_cached_content(cached_content)
        return cached_context_model

    @GoogleBaseHook.fallback_to_default_project_id
    def run_evaluation(
        self,
        pretrained_model: str,
        eval_dataset: dict,
        metrics: list,
        experiment_name: str,
        experiment_run_name: str,
        prompt_template: str,
        location: str,
        generation_config: dict | None = None,
        safety_settings: dict | None = None,
        system_instruction: str | None = None,
        tools: list | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
    ) -> EvalResult:
        """
        Use the Rapid Evaluation API to evaluate a model.

        :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
        :param location: Required. The ID of the Google Cloud location that the service belongs to.
        :param pretrained_model: Required. A pre-trained model optimized for performing natural
            language tasks such as classification, summarization, extraction, content
            creation, and ideation.
        :param eval_dataset: Required. A fixed dataset for evaluating a model against. Adheres to Rapid Evaluation API.
        :param metrics: Required. A list of evaluation metrics to be used in the experiment. Adheres to Rapid Evaluation API.
        :param experiment_name: Required. The name of the evaluation experiment.
        :param experiment_run_name: Required. The specific run name or ID for this experiment.
        :param prompt_template: Required. The template used to format the model's prompts during evaluation. Adheres to Rapid Evaluation API.
        :param generation_config: Optional. A dictionary containing generation parameters for the model.
        :param safety_settings: Optional. A dictionary specifying harm category thresholds for blocking model outputs.
        :param system_instruction: Optional. An instruction given to the model to guide its behavior.
        :param tools: Optional. A list of tools available to the model during evaluation, such as a data store.
        """
        vertexai.init(project=project_id, location=location, credentials=self.get_credentials())

        model = self.get_generative_model(
            pretrained_model=pretrained_model,
            system_instruction=system_instruction,
            generation_config=generation_config,
            safety_settings=safety_settings,
            tools=tools,
        )

        eval_task = self.get_eval_task(
            dataset=eval_dataset,
            metrics=metrics,
            experiment=experiment_name,
        )

        eval_result = eval_task.evaluate(
            model=model,
            prompt_template=prompt_template,
            experiment_run_name=experiment_run_name,
        )

        return eval_result



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
System test for ComputeEngineInsertInstanceOperator
verifying recreate_if_machine_type_different=True recreates the
correct machine_type instance when machine_type drifts.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.operators.python import PythonOperator
from airflow.providers.google.cloud.hooks.compute import ComputeEngineHook
from airflow.providers.google.cloud.operators.compute import (
    ComputeEngineDeleteInstanceOperator,
    ComputeEngineInsertInstanceOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID

DAG_ID = "cloud_compute_insert_recreate_if_different"
LOCATION = "us-central1-a"

INSTANCE_NAME = f"airflow-drift-test-{ENV_ID}"
MACHINE_TYPE_A = "n1-standard-1"
MACHINE_TYPE_B = "n1-standard-2"

BASE_BODY = {
    "name": INSTANCE_NAME,
    "disks": [
        {
            "boot": True,
            "auto_delete": True,
            "initialize_params": {
                "disk_size_gb": "10",
                "source_image": "projects/debian-cloud/global/images/family/debian-12",
            },
        }
    ],
    "network_interfaces": [{"network": "global/networks/default"}],
}


def assert_machine_type():
    hook = ComputeEngineHook()
    instance = hook.get_instance(
        project_id=PROJECT_ID,
        zone=LOCATION,
        resource_id=INSTANCE_NAME,
    )

    machine_type = instance.machine_type.split("/")[-1]

    assert machine_type == MACHINE_TYPE_B, f"Expected machine type {MACHINE_TYPE_B}, got {machine_type}"


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "compute"],
) as dag:
    # Step 1: Create with machine type A.
    create_instance = ComputeEngineInsertInstanceOperator(
        task_id="create_instance",
        project_id=PROJECT_ID,
        zone=LOCATION,
        body={
            **BASE_BODY,
            "machine_type": f"zones/{LOCATION}/machineTypes/{MACHINE_TYPE_A}",
        },
    )

    # Step 2: Re-run with different machine type and recreate recreate_if_machine_type_different=True.
    recreate_instance = ComputeEngineInsertInstanceOperator(
        task_id="recreate_instance",
        project_id=PROJECT_ID,
        zone=LOCATION,
        body={
            **BASE_BODY,
            "machine_type": f"zones/{LOCATION}/machineTypes/{MACHINE_TYPE_B}",
        },
        recreate_if_machine_type_different=True,
    )

    # Step 3: Validate new machine type.
    validate_machine_type = PythonOperator(
        task_id="validate_machine_type",
        python_callable=assert_machine_type,
    )

    # Step 4: Cleanup.
    delete_instance = ComputeEngineDeleteInstanceOperator(
        task_id="delete_instance",
        project_id=PROJECT_ID,
        zone=LOCATION,
        resource_id=INSTANCE_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    create_instance >> recreate_instance >> validate_machine_type >> delete_instance

    # Everything below this line is required for system tests.
    from tests_common.test_utils.watcher import watcher

    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest.mock import MagicMock, patch

import pytest

pytest.importorskip("gcsfs")

TEST_CONN = "google_cloud_test_conn"


@pytest.fixture(scope="module", autouse=True)
def _setup_connections():
    with pytest.MonkeyPatch.context() as mp_ctx:
        mp_ctx.setenv(f"AIRFLOW_CONN_{TEST_CONN}".upper(), "google-cloud-platform://")
        yield


class TestGCSFilesystem:
    @patch("airflow.providers.google.cloud.fs.gcs.GoogleBaseHook")
    @patch("gcsfs.GCSFileSystem")
    def test_get_fs_passes_credentials_object(self, mock_gcsfs, mock_hook):
        """Test that get_fs passes a Credentials object to GCSFileSystem."""
        from airflow.providers.google.cloud.fs.gcs import get_fs

        mock_credentials = MagicMock()
        mock_hook_instance = MagicMock()
        mock_hook_instance.get_credentials.return_value = mock_credentials
        mock_hook_instance.project_id = "test-project"
        mock_hook_instance.extras = {}
        mock_hook.return_value = mock_hook_instance

        get_fs(conn_id=TEST_CONN)

        mock_hook_instance.get_credentials.assert_called_once()
        call_kwargs = mock_gcsfs.call_args.kwargs
        assert call_kwargs["token"] is mock_credentials

    @patch("gcsfs.GCSFileSystem")
    def test_get_fs_no_conn_id(self, mock_gcsfs):
        """Test that get_fs works without conn_id."""
        from airflow.providers.google.cloud.fs.gcs import get_fs

        get_fs(conn_id=None)

        mock_gcsfs.assert_called_once_with()

    @patch("airflow.providers.google.cloud.fs.gcs.GoogleBaseHook")
    @patch("gcsfs.GCSFileSystem")
    def test_get_fs_with_anonymous_credentials(self, mock_gcsfs, mock_hook):
        """Test that get_fs works with anonymous credentials."""
        from google.auth.credentials import AnonymousCredentials

        from airflow.providers.google.cloud.fs.gcs import get_fs

        anonymous_creds = AnonymousCredentials()
        mock_hook_instance = MagicMock()
        mock_hook_instance.get_credentials.return_value = anonymous_creds
        mock_hook_instance.project_id = None
        mock_hook_instance.extras = {}
        mock_hook.return_value = mock_hook_instance

        get_fs(conn_id=TEST_CONN)

        call_kwargs = mock_gcsfs.call_args.kwargs
        assert isinstance(call_kwargs["token"], AnonymousCredentials)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

from google.cloud.bigquery import CopyJob, DatasetReference, ExtractJob, LoadJob, QueryJob, TableReference

from airflow.providers.common.compat.assets import Asset
from airflow.providers.google.cloud.utils.lineage import (
    _add_bq_table_to_lineage,
    _add_gcs_uris_to_lineage,
    send_hook_lineage_for_bq_job,
)

PROJECT_ID = "test-project"
DATASET_ID = "test_dataset"
TABLE_ID = "test_table"
JOB_ID = "test-job-123"

TABLE_REFERENCE = TableReference(DatasetReference(PROJECT_ID, DATASET_ID), TABLE_ID)


def _make_table_ref(project, dataset, table):
    return TableReference(DatasetReference(project, dataset), table)


class TestAddBqTableToLineage:
    def test_add_as_input(self):
        collector = mock.MagicMock()
        context = mock.sentinel.context

        _add_bq_table_to_lineage(collector, context, TABLE_REFERENCE, is_input=True)

        collector.add_input_asset.assert_called_once_with(
            context=context,
            scheme="bigquery",
            asset_kwargs={
                "project_id": PROJECT_ID,
                "dataset_id": DATASET_ID,
                "table_id": TABLE_ID,
            },
        )
        collector.add_output_asset.assert_not_called()

    def test_add_as_output(self):
        collector = mock.MagicMock()
        context = mock.sentinel.context

        _add_bq_table_to_lineage(collector, context, TABLE_REFERENCE, is_input=False)

        collector.add_output_asset.assert_called_once_with(
            context=context,
            scheme="bigquery",
            asset_kwargs={
                "project_id": PROJECT_ID,
                "dataset_id": DATASET_ID,
                "table_id": TABLE_ID,
            },
        )
        collector.add_input_asset.assert_not_called()


class TestAddGcsUrisToLineage:
    def test_add_uris_as_input(self):
        collector = mock.MagicMock()
        context = mock.sentinel.context
        uris = ["gs://bucket1/path/file.csv", "gs://bucket2/other.json"]

        _add_gcs_uris_to_lineage(collector, context, uris, is_input=True)

        assert collector.add_input_asset.call_count == 2
        collector.add_input_asset.assert_any_call(context=context, uri="gs://bucket1/path/file.csv")
        collector.add_input_asset.assert_any_call(context=context, uri="gs://bucket2/other.json")
        collector.add_output_asset.assert_not_called()

    def test_add_uris_as_output(self):
        collector = mock.MagicMock()
        context = mock.sentinel.context
        uris = ["gs://bucket/export/data.csv"]

        _add_gcs_uris_to_lineage(collector, context, uris, is_input=False)

        collector.add_output_asset.assert_called_once_with(context=context, uri="gs://bucket/export/data.csv")
        collector.add_input_asset.assert_not_called()

    def test_empty_uris(self):
        collector = mock.MagicMock()
        _add_gcs_uris_to_lineage(collector, mock.sentinel.context, [], is_input=True)
        collector.add_input_asset.assert_not_called()

    def test_none_uris(self):
        collector = mock.MagicMock()
        _add_gcs_uris_to_lineage(collector, mock.sentinel.context, None, is_input=True)
        collector.add_input_asset.assert_not_called()


class TestSendHookLineageForBqJob:
    @mock.patch("airflow.providers.google.cloud.utils.lineage.send_sql_hook_lineage")
    def test_query_job(self, mock_send_sql):
        job = mock.MagicMock(spec=QueryJob)
        job.query = "SELECT * FROM dataset.table"
        job.job_id = JOB_ID
        job.default_dataset = DatasetReference(PROJECT_ID, DATASET_ID)
        context = mock.sentinel.context

        send_hook_lineage_for_bq_job(context=context, job=job)

        mock_send_sql.assert_called_once_with(
            context=context,
            sql="SELECT * FROM dataset.table",
            job_id=JOB_ID,
            default_db=PROJECT_ID,
            default_schema=DATASET_ID,
        )

    @mock.patch("airflow.providers.google.cloud.utils.lineage.send_sql_hook_lineage")
    def test_query_job_no_default_dataset(self, mock_send_sql):
        job = mock.MagicMock(spec=QueryJob)
        job.query = "SELECT 1"
        job.job_id = JOB_ID
        job.default_dataset = None
        context = mock.sentinel.context

        send_hook_lineage_for_bq_job(context=context, job=job)

        mock_send_sql.assert_called_once_with(
            context=context,
            sql="SELECT 1",
            job_id=JOB_ID,
            default_db=None,
            default_schema=None,
        )

    def test_load_job(self, hook_lineage_collector):
        job = mock.MagicMock(spec=LoadJob)
        job.source_uris = ["gs://bucket/data.csv", "gs://bucket/data2.csv"]
        job.destination = TABLE_REFERENCE
        context = mock.sentinel.context

        send_hook_lineage_for_bq_job(context=context, job=job)

        assert len(hook_lineage_collector.collected_assets.inputs) == 2
        assert len(hook_lineage_collector.collected_assets.outputs) == 1
        assert hook_lineage_collector.collected_assets.outputs[0].asset == Asset(
            uri=f"bigquery://{PROJECT_ID}/{DATASET_ID}/{TABLE_ID}"
        )

    def test_load_job_no_destination(self, hook_lineage_collector):
        job = mock.MagicMock(spec=LoadJob)
        job.source_uris = ["gs://bucket/data.csv"]
        job.destination = None
        context = mock.sentinel.context

        send_hook_lineage_for_bq_job(context=context, job=job)

        assert len(hook_lineage_collector.collected_assets.inputs) == 1
        assert len(hook_lineage_collector.collected_assets.outputs) == 0

    def test_copy_job(self, hook_lineage_collector):
        source1 = _make_table_ref(PROJECT_ID, DATASET_ID, "source1")
        source2 = _make_table_ref(PROJECT_ID, DATASET_ID, "source2")
        dest = _make_table_ref(PROJECT_ID, DATASET_ID, "dest")

        job = mock.MagicMock(spec=CopyJob)
        job.sources = [source1, source2]
        job.destination = dest
        context = mock.sentinel.context

        send_hook_lineage_for_bq_job(context=context, job=job)

        assert len(hook_lineage_collector.collected_assets.inputs) == 2
        assert len(hook_lineage_collector.collected_assets.outputs) == 1
        assert hook_lineage_collector.collected_assets.inputs[0].asset == Asset(
            uri=f"bigquery://{PROJECT_ID}/{DATASET_ID}/source1"
        )
        assert hook_lineage_collector.collected_assets.inputs[1].asset == Asset(
            uri=f"bigquery://{PROJECT_ID}/{DATASET_ID}/source2"
        )
        assert hook_lineage_collector.collected_assets.outputs[0].asset == Asset(
            uri=f"bigquery://{PROJECT_ID}/{DATASET_ID}/dest"
        )

    def test_extract_job(self, hook_lineage_collector):
        job = mock.MagicMock(spec=ExtractJob)
        job.source = TABLE_REFERENCE
        job.destination_uris = ["gs://bucket/export/file1.csv", "gs://bucket/export/file2.csv"]
        context = mock.sentinel.context

        send_hook_lineage_for_bq_job(context=context, job=job)

        assert len(hook_lineage_collector.collected_assets.inputs) == 1
        assert len(hook_lineage_collector.collected_assets.outputs) == 2
        assert hook_lineage_collector.collected_assets.inputs[0].asset == Asset(
            uri=f"bigquery://{PROJECT_ID}/{DATASET_ID}/{TABLE_ID}"
        )

    def test_extract_job_no_source(self, hook_lineage_collector):
        job = mock.MagicMock(spec=ExtractJob)
        job.source = None
        job.destination_uris = ["gs://bucket/export/file.csv"]
        context = mock.sentinel.context

        send_hook_lineage_for_bq_job(context=context, job=job)

        assert len(hook_lineage_collector.collected_assets.inputs) == 0
        assert len(hook_lineage_collector.collected_assets.outputs) == 1

    @mock.patch("airflow.providers.google.cloud.utils.lineage.send_sql_hook_lineage")
    def test_unknown_job_type_does_not_raise(self, mock_send_sql, hook_lineage_collector):
        job = mock.MagicMock()
        send_hook_lineage_for_bq_job(context=mock.sentinel.context, job=job)
        mock_send_sql.assert_not_called()
        assert len(hook_lineage_collector.collected_assets.inputs) == 0
        assert len(hook_lineage_collector.collected_assets.outputs) == 0

    def test_exception_in_non_query_job_is_caught(self, hook_lineage_collector):
        job = mock.MagicMock(spec=LoadJob)
        type(job).source_uris = mock.PropertyMock(side_effect=RuntimeError("boom"))
        context = mock.sentinel.context

        send_hook_lineage_for_bq_job(context=context, job=job)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG that demonstrates using Google Cloud Pub/Sub with MessageQueueTrigger
and Asset Watchers for event-driven workflows.

This example shows how to create a DAG that triggers when messages arrive in a
Google Cloud Pub/Sub subscription using Asset Watchers.

Prerequisites
-------------

Before running this example, ensure you have:

1. A GCP project with Pub/Sub API enabled
2. The following Pub/Sub resources created in your project:

   - Topic: ``test-topic``
   - Subscription: ``test-subscription``

You can create these resources using:

.. code-block:: bash

    # Create topic
    gcloud pubsub topics create test-topic --project={PROJECT_ID}

    # Create subscription
    gcloud pubsub subscriptions create test-subscription \\
        --topic=test-topic --project={PROJECT_ID}

How to test
-----------

1. Ensure the Pub/Sub resources exist (see Prerequisites above)
2. Publish a message to trigger the DAG:

   .. code-block:: bash

       gcloud pubsub topics publish test-topic \\
           --message="Test message" --project={PROJECT_ID}

3. The DAG will be triggered automatically when the message arrives
"""

from __future__ import annotations

# [START howto_trigger_pubsub_message_queue]
from airflow.providers.common.messaging.triggers.msg_queue import MessageQueueTrigger
from airflow.providers.standard.operators.empty import EmptyOperator
from airflow.sdk import DAG, Asset, AssetWatcher

# Define a trigger that listens to a Google Cloud Pub/Sub subscription
trigger = MessageQueueTrigger(
    scheme="google+pubsub",
    project_id="my-project",
    subscription="test-subscription",
    ack_messages=True,
    max_messages=1,
    gcp_conn_id="google_cloud_default",
    poke_interval=60.0,
)

# Define an asset that watches for messages on the Pub/Sub subscription
asset = Asset(
    "event_schedule_pubsub_asset_1",
    watchers=[AssetWatcher(name="event_schedule_pubsub_watcher_1", trigger=trigger)],
)

with DAG(
    dag_id="example_event_schedule_pubsub",
    schedule=[asset],
) as dag:
    process_message_task = EmptyOperator(task_id="process_pubsub_message")
# [END howto_trigger_pubsub_message_queue]


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import asyncio
import logging
from unittest import mock

import pytest

from airflow.models import Connection
from airflow.providers.google.cloud.hooks.gen_ai import BatchJobStatus
from airflow.providers.google.cloud.triggers.gen_ai import (
    GenAIGeminiCreateBatchJobTrigger,
    GenAIGeminiCreateEmbeddingsBatchJobTrigger,
)
from airflow.triggers.base import TriggerEvent

BATCH_JOB_INLINED_REQUESTS = [
    {"contents": [{"parts": [{"text": "Tell me a one-sentence joke."}], "role": "user"}]},
    {"contents": [{"parts": [{"text": "Why is the sky blue?"}], "role": "user"}]},
]

EMBEDDINGS_JOB_INLINED_REQUESTS = {
    "contents": [{"parts": [{"text": "Why is the sky blue?"}], "role": "user"}]
}
GEMINI_API_KEY = "test-key"
GEMINI_MODEL = "test-gemini-model"
BATCH_JOB_NAME = "test-name"
FILE_NAME = "test-file"
FILE_PATH = "test/path/to/file"
TASK_ID = "test_task_id"
GCP_PROJECT = "test-project"
GCP_LOCATION = "test-location"
GCP_CONN_ID = "test-conn"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
BATCH_JOB_CONFIG = {"display_name": "test-batch-job"}
RESULTS_FOLDER = "test/results/folder"


@pytest.fixture
@mock.patch(
    "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_connection",
    return_value=Connection(conn_id=GCP_CONN_ID),
)
def create_batch_job_trigger(mock_conn):
    return GenAIGeminiCreateBatchJobTrigger(
        project_id=GCP_PROJECT,
        location=GCP_LOCATION,
        gcp_conn_id=GCP_CONN_ID,
        impersonation_chain=IMPERSONATION_CHAIN,
        model=GEMINI_MODEL,
        input_source=BATCH_JOB_INLINED_REQUESTS,
        create_batch_job_config=BATCH_JOB_CONFIG,
        gemini_api_key=GEMINI_API_KEY,
        retrieve_result=True,
        polling_interval=30,
        results_folder=RESULTS_FOLDER,
    )


@pytest.fixture
@mock.patch(
    "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.get_connection",
    return_value=Connection(conn_id=GCP_CONN_ID),
)
def create_embeddings_batch_job_trigger(mock_conn):
    return GenAIGeminiCreateEmbeddingsBatchJobTrigger(
        project_id=GCP_PROJECT,
        location=GCP_LOCATION,
        gcp_conn_id=GCP_CONN_ID,
        impersonation_chain=IMPERSONATION_CHAIN,
        model=GEMINI_MODEL,
        input_source=BATCH_JOB_INLINED_REQUESTS,
        create_embeddings_config=BATCH_JOB_CONFIG,
        gemini_api_key=GEMINI_API_KEY,
        retrieve_result=True,
        polling_interval=30,
        results_folder=RESULTS_FOLDER,
    )


class TestGenAIGeminiCreateBatchJobTrigger:
    def test_serialize(self, create_batch_job_trigger):
        actual_data = create_batch_job_trigger.serialize()
        expected_data = (
            "airflow.providers.google.cloud.triggers.gen_ai.GenAIGeminiCreateBatchJobTrigger",
            {
                "project_id": GCP_PROJECT,
                "location": GCP_LOCATION,
                "gcp_conn_id": GCP_CONN_ID,
                "impersonation_chain": IMPERSONATION_CHAIN,
                "model": GEMINI_MODEL,
                "input_source": BATCH_JOB_INLINED_REQUESTS,
                "create_batch_job_config": BATCH_JOB_CONFIG,
                "gemini_api_key": GEMINI_API_KEY,
                "retrieve_result": True,
                "polling_interval": 30,
                "results_folder": RESULTS_FOLDER,
            },
        )
        assert actual_data == expected_data

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.create_batch_job")
    @mock.patch("airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.get_batch_job")
    async def test_run_loop_return_success_event(
        self, mock_job_status, mock_create_batch_job, create_batch_job_trigger
    ):
        mock_job_status.return_value.state.name = BatchJobStatus.SUCCEEDED.value
        mock_job_status.return_value.name = BATCH_JOB_CONFIG["display_name"]

        expected_event = TriggerEvent(
            {
                "status": "success",
                "message": "Job completed",
                "job_name": BATCH_JOB_CONFIG["display_name"],
            }
        )
        actual_event = await create_batch_job_trigger.run().asend(None)

        mock_create_batch_job.assert_called_once()
        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.create_batch_job")
    @mock.patch("airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.get_batch_job")
    async def test_run_loop_return_failed_event(
        self, mock_job_status, mock_create_batch_job, create_batch_job_trigger
    ):
        mock_job_status.return_value.state.name = BatchJobStatus.FAILED.value
        mock_job_status.return_value.name = BATCH_JOB_NAME

        expected_event = TriggerEvent(
            {"status": "error", "message": f"Job {BATCH_JOB_NAME} execution was not completed!"}
        )
        actual_event = await create_batch_job_trigger.run().asend(None)

        mock_create_batch_job.assert_called_once()
        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch("airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.create_batch_job")
    @mock.patch("airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.get_batch_job")
    async def test_run_loop_is_still_running(
        self, mock_job_status, mock_create_batch_job, create_batch_job_trigger, caplog
    ):
        mock_job_status.return_value.state.name = BatchJobStatus.RUNNING.value
        caplog.set_level(logging.INFO)

        task = asyncio.create_task(create_batch_job_trigger.run().__anext__())
        await asyncio.sleep(5)

        assert not task.done()
        assert (
            f"Waiting for job execution, polling interval: 30 seconds, current state: {BatchJobStatus.RUNNING.value}"
            in caplog.text
        )
        # cancel the task to suppress test warnings
        task.cancel()


class TestGenAIGeminiCreateEmbeddingsBatchJobTrigger:
    def test_serialize(self, create_embeddings_batch_job_trigger):
        actual_data = create_embeddings_batch_job_trigger.serialize()
        expected_data = (
            "airflow.providers.google.cloud.triggers.gen_ai.GenAIGeminiCreateEmbeddingsBatchJobTrigger",
            {
                "project_id": GCP_PROJECT,
                "location": GCP_LOCATION,
                "gcp_conn_id": GCP_CONN_ID,
                "impersonation_chain": IMPERSONATION_CHAIN,
                "model": GEMINI_MODEL,
                "input_source": BATCH_JOB_INLINED_REQUESTS,
                "create_embeddings_config": BATCH_JOB_CONFIG,
                "gemini_api_key": GEMINI_API_KEY,
                "retrieve_result": True,
                "polling_interval": 30,
                "results_folder": RESULTS_FOLDER,
            },
        )
        assert actual_data == expected_data

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.create_embeddings_batch_job"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.get_batch_job")
    async def test_run_loop_return_success_event(
        self, mock_job_status, mock_create_embeddings_batch_job, create_embeddings_batch_job_trigger
    ):
        test_job_model_dump = {"id": "test_job_id", "status": "succeeded"}
        mock_job_status.return_value.state.name = BatchJobStatus.SUCCEEDED.value
        mock_job_status.return_value.model_dump = mock.Mock(return_value=test_job_model_dump)

        expected_event = TriggerEvent(
            {
                "status": "success",
                "message": "Job completed",
                "job": test_job_model_dump,
            }
        )
        actual_event = await create_embeddings_batch_job_trigger.run().asend(None)
        mock_create_embeddings_batch_job.assert_called_once()
        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.create_embeddings_batch_job"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.get_batch_job")
    async def test_run_loop_return_failed_event(
        self, mock_job_status, mock_create_embeddings_batch_job, create_embeddings_batch_job_trigger
    ):
        mock_job_status.return_value.state.name = BatchJobStatus.FAILED.value
        mock_job_status.return_value.name = BATCH_JOB_NAME

        expected_event = TriggerEvent(
            {"status": "error", "message": f"Job {BATCH_JOB_NAME} execution was not completed!"}
        )
        actual_event = await create_embeddings_batch_job_trigger.run().asend(None)

        mock_create_embeddings_batch_job.assert_called_once()
        assert actual_event == expected_event

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.create_embeddings_batch_job"
    )
    @mock.patch("airflow.providers.google.cloud.hooks.gen_ai.GenAIGeminiAPIAsyncHook.get_batch_job")
    async def test_run_loop_is_still_running(
        self, mock_job_status, mock_create_embeddings_batch_job, create_embeddings_batch_job_trigger, caplog
    ):
        mock_job_status.return_value.state.name = BatchJobStatus.RUNNING.value
        caplog.set_level(logging.INFO)

        task = asyncio.create_task(create_embeddings_batch_job_trigger.run().__anext__())
        await asyncio.sleep(5)

        assert not task.done()
        assert (
            f"Waiting for job execution, polling interval: 30 seconds, current state: {BatchJobStatus.RUNNING.value}"
            in caplog.text
        )
        # cancel the task to suppress test warnings
        task.cancel()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from ray.dashboard.modules.job.common import JobStatus

from airflow.providers.common.compat.sdk import AirflowTaskTimeout
from airflow.providers.google.cloud.operators.ray import (
    RayDeleteJobOperator,
    RayGetJobInfoOperator,
    RayListJobsOperator,
    RayStopJobOperator,
    RaySubmitJobOperator,
)

TASK_ID = "test-task"
GCP_CONN_ID = "test-gcp-conn-id"
IMPERSONATION_CHAIN = "test-impersonation"
CLUSTER_ADDRESS = "ray-head-123.us-central1.ray.googleusercontent.com"
ENTRYPOINT = "python3 heavy.py"
SUBMISSION_ID = "submission-123"
JOB_ID = "job-123"

RAY_OP_PATH = "airflow.providers.google.cloud.operators.ray.{}"


class TestRaySubmitJobOperator:
    @mock.patch(RAY_OP_PATH.format("RayJobHook"))
    def test_execute_submits_job_and_persists_link(self, mock_hook_cls):
        mock_hook = mock_hook_cls.return_value
        mock_hook.submit_job.return_value = JOB_ID

        op = RaySubmitJobOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            cluster_address=CLUSTER_ADDRESS,
            entrypoint=ENTRYPOINT,
            wait_for_job_done=False,
            get_job_logs=False,
            runtime_env={"k": "v"},
            metadata={"m": "v"},
            submission_id=SUBMISSION_ID,
            entrypoint_num_cpus=1,
            entrypoint_num_gpus=0,
            entrypoint_memory=1024,
            entrypoint_resources={"CPU": 1.0},
        )

        ti_mock = mock.MagicMock()
        context = {"ti": ti_mock, "task": mock.MagicMock()}

        op.execute(context=context)

        mock_hook_cls.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        mock_hook.submit_job.assert_called_once_with(
            cluster_address=CLUSTER_ADDRESS,
            entrypoint=ENTRYPOINT,
            runtime_env={"k": "v"},
            metadata={"m": "v"},
            submission_id=SUBMISSION_ID,
            entrypoint_num_cpus=1,
            entrypoint_num_gpus=0,
            entrypoint_memory=1024,
            entrypoint_resources={"CPU": 1.0},
        )
        ti_mock.xcom_push.assert_called_once_with(
            key="ray_job",
            value={
                "cluster_address": CLUSTER_ADDRESS,
                "job_id": JOB_ID,
            },
        )
        mock_hook.get_job_status.assert_not_called()
        mock_hook.get_job_logs.assert_not_called()

    def test_execute_raises_if_logs_without_wait(self):
        op = RaySubmitJobOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            cluster_address=CLUSTER_ADDRESS,
            entrypoint=ENTRYPOINT,
            wait_for_job_done=False,
            get_job_logs=True,
        )

        ti_mock = mock.MagicMock()
        context = {"ti": ti_mock, "task": mock.MagicMock()}

        with pytest.raises(ValueError, match="Retrieving Job logs can be possible only after Job completion"):
            op.execute(context=context)

    @mock.patch(RAY_OP_PATH.format("RayJobHook"))
    def test_execute_waits_for_job_and_gets_logs(self, mock_hook_cls):
        mock_hook = mock_hook_cls.return_value
        mock_hook.submit_job.return_value = JOB_ID
        mock_hook.get_job_status.return_value = JobStatus.SUCCEEDED
        mock_hook.get_job_logs.return_value = "some logs"

        op = RaySubmitJobOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            cluster_address=CLUSTER_ADDRESS,
            entrypoint=ENTRYPOINT,
            wait_for_job_done=True,
            get_job_logs=True,
        )

        ti_mock = mock.MagicMock()
        context = {"ti": ti_mock, "task": mock.MagicMock()}

        op.execute(context=context)
        mock_hook.submit_job.assert_called_once_with(
            cluster_address=CLUSTER_ADDRESS,
            entrypoint=ENTRYPOINT,
            runtime_env=None,
            metadata=None,
            submission_id=None,
            entrypoint_num_cpus=None,
            entrypoint_num_gpus=None,
            entrypoint_memory=None,
            entrypoint_resources=None,
        )
        mock_hook.get_job_status.assert_called_with(
            cluster_address=CLUSTER_ADDRESS,
            job_id=JOB_ID,
        )

        mock_hook.get_job_logs.assert_called_once_with(
            cluster_address=CLUSTER_ADDRESS,
            job_id=JOB_ID,
        )

        ti_mock.xcom_push.assert_called_once_with(
            key="ray_job",
            value={
                "cluster_address": CLUSTER_ADDRESS,
                "job_id": JOB_ID,
            },
        )

    @mock.patch(RAY_OP_PATH.format("time.sleep"))
    @mock.patch(RAY_OP_PATH.format("RayJobHook"))
    def test_check_job_status_reaches_terminal(self, mock_hook_cls, mock_sleep):
        mock_hook = mock_hook_cls.return_value
        mock_hook.stop_job.return_value = True

        operator = RaySubmitJobOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            cluster_address=CLUSTER_ADDRESS,
            entrypoint=ENTRYPOINT,
            wait_for_job_done=True,
            get_job_logs=True,
        )
        operator.hook.get_job_status = mock.MagicMock(
            side_effect=[JobStatus.RUNNING, JobStatus.RUNNING, JobStatus.SUCCEEDED]
        )
        status = operator._check_job_status("addr", "job", polling_interval=1, timeout=100)

        assert status == JobStatus.SUCCEEDED
        assert mock_sleep.call_count == 2

    @mock.patch(RAY_OP_PATH.format("time.sleep"))
    @mock.patch(RAY_OP_PATH.format("time.monotonic"))
    @mock.patch(RAY_OP_PATH.format("RayJobHook"))
    def test_check_job_status_timeout(self, mock_hook_cls, mock_monotonic, mock_sleep):
        mock_hook = mock_hook_cls.return_value
        mock_hook.stop_job.return_value = True
        mock_monotonic.side_effect = [0, 10, 20, 1000]
        operator = RaySubmitJobOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            cluster_address=CLUSTER_ADDRESS,
            entrypoint=ENTRYPOINT,
            wait_for_job_done=True,
            get_job_logs=True,
        )
        operator.hook.get_job_status = mock.MagicMock(return_value=JobStatus.RUNNING)

        with pytest.raises(
            AirflowTaskTimeout, match=r"Timeout waiting for Ray Job job to finish. Last status: RUNNING"
        ):
            operator._check_job_status("addr", "job", polling_interval=1, timeout=30)


class TestRayStopJobOperator:
    @mock.patch(RAY_OP_PATH.format("RayJobHook"))
    def test_execute_stops_job_successfully(self, mock_hook_cls):
        mock_hook = mock_hook_cls.return_value
        mock_hook.stop_job.return_value = True

        op = RayStopJobOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            cluster_address=CLUSTER_ADDRESS,
            job_id=JOB_ID,
        )

        context = {"ti": mock.MagicMock(), "task": mock.MagicMock()}
        result = op.execute(context=context)

        mock_hook_cls.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.stop_job.assert_called_once_with(
            cluster_address=CLUSTER_ADDRESS,
            job_id=JOB_ID,
        )
        assert result is None


class TestRayDeleteJobOperator:
    @mock.patch(RAY_OP_PATH.format("RayJobHook"))
    def test_execute_deletes_job_successfully(self, mock_hook_cls):
        mock_hook = mock_hook_cls.return_value
        mock_hook.delete_job.return_value = True

        op = RayDeleteJobOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            cluster_address=CLUSTER_ADDRESS,
            job_id=JOB_ID,
        )

        context = {"ti": mock.MagicMock(), "task": mock.MagicMock()}
        result = op.execute(context=context)

        mock_hook_cls.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.delete_job.assert_called_once_with(
            cluster_address=CLUSTER_ADDRESS,
            job_id=JOB_ID,
        )
        assert result is None


class TestRayGetJobInfoOperator:
    @mock.patch(RAY_OP_PATH.format("RayJobHook"))
    def test_execute_returns_serialized_job_info(self, mock_hook_cls):
        mock_hook = mock_hook_cls.return_value
        fake_job_obj = object()
        fake_serialized = {"job_id": JOB_ID, "status": "SUCCEEDED"}

        mock_hook.get_job_info.return_value = fake_job_obj
        mock_hook.serialize_job_obj.return_value = fake_serialized

        op = RayGetJobInfoOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            cluster_address=CLUSTER_ADDRESS,
            job_id=JOB_ID,
        )

        context = {"ti": mock.MagicMock(), "task": mock.MagicMock()}
        result = op.execute(context=context)

        mock_hook_cls.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.get_job_info.assert_called_once_with(
            cluster_address=CLUSTER_ADDRESS,
            job_id=JOB_ID,
        )
        mock_hook.serialize_job_obj.assert_called_once_with(fake_job_obj)
        assert result == fake_serialized


class TestRayListJobsOperator:
    @mock.patch(RAY_OP_PATH.format("RayJobHook"))
    def test_execute_lists_and_serializes_jobs(self, mock_hook_cls):
        mock_hook = mock_hook_cls.return_value
        job1 = object()
        job2 = object()
        mock_hook.list_jobs.return_value = [job1, job2]
        mock_hook.serialize_job_obj.side_effect = [
            {"job_id": "job-1"},
            {"job_id": "job-2"},
        ]

        op = RayListJobsOperator(
            task_id=TASK_ID,
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
            cluster_address=CLUSTER_ADDRESS,
        )

        context = {"ti": mock.MagicMock(), "task": mock.MagicMock()}
        result = op.execute(context=context)

        mock_hook_cls.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )
        mock_hook.list_jobs.assert_called_once_with(
            cluster_address=CLUSTER_ADDRESS,
        )
        mock_hook.serialize_job_obj.assert_has_calls([mock.call(job1), mock.call(job2)])
        assert result == [
            {"job_id": "job-1"},
            {"job_id": "job-2"},
        ]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from airflow.providers.google.cloud.links.base import BaseGoogleLink

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context

RAY_JOB_BASE_LINK = "http://{cluster_address}/#/jobs"
RAY_JOB_LINK = RAY_JOB_BASE_LINK + "/{job_id}"


class RayJobLink(BaseGoogleLink):
    """Helper class for constructing Job on Ray cluster Link."""

    name = "Ray Job"
    key = "ray_job"
    format_str = RAY_JOB_LINK

    @classmethod
    def persist(cls, context: Context, **value):
        cluster_address = value.get("cluster_address")
        job_id = value.get("job_id")
        super().persist(
            context=context,
            cluster_address=cluster_address,
            job_id=job_id,
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Real tests for Google Cloud Storage Transfer Service links."""

from __future__ import annotations

import pytest

from airflow.providers.google.cloud.links.cloud_storage_transfer import (
    CloudStorageTransferDetailsLink,
    CloudStorageTransferJobLink,
    CloudStorageTransferLinkHelper,
    CloudStorageTransferListLink,
)

REAL_PROJECT_ID = "my-gcp-project-123456"
REAL_TRANSFER_JOB = "transferJobs-1234567890123456789"
REAL_TRANSFER_OPERATION = "transferOperations/9876543210987654321"
REAL_OPERATION_NAME = f"{REAL_TRANSFER_OPERATION}-{REAL_TRANSFER_JOB}"

EXPECTED_LIST_URL = f"https://console.cloud.google.com/transfer/jobs?project={REAL_PROJECT_ID}"
EXPECTED_JOB_URL = f"https://console.cloud.google.com/transfer/jobs/transferJobs%2F{REAL_TRANSFER_JOB}/runs?project={REAL_PROJECT_ID}"
EXPECTED_DETAILS_URL = (
    f"https://console.cloud.google.com/transfer/jobs/transferJobs%2FtransferJobs"
    f"/runs/transferOperations%2F9876543210987654321-transferJobs-1234567890123456789?project={REAL_PROJECT_ID}"
)


class TestCloudStorageTransferLinkHelper:
    """Test the CloudStorageTransferLinkHelper with real operation names."""

    def test_extract_parts_with_real_operation_name(self):
        """Test extract_parts with a real Google Cloud Storage Transfer operation name."""
        transfer_operation, transfer_job = CloudStorageTransferLinkHelper.extract_parts(REAL_OPERATION_NAME)

        assert transfer_operation == "9876543210987654321-transferJobs-1234567890123456789"
        assert transfer_job == "transferJobs"

    def test_extract_parts_with_various_real_formats(self):
        """Test extract_parts with various real operation name formats."""
        test_cases = [
            ("transferOperations/12345-transferJobs-67890", "12345-transferJobs-67890", "transferJobs"),
            ("transferOperations/op123-transferJobs-job456", "op123-transferJobs-job456", "transferJobs"),
            (
                "transferOperations/99999999999999999999-transferJobs-11111111111111111111",
                "99999999999999999999-transferJobs-11111111111111111111",
                "transferJobs",
            ),
        ]

        for operation_name, expected_operation, expected_job in test_cases:
            transfer_operation, transfer_job = CloudStorageTransferLinkHelper.extract_parts(operation_name)
            assert transfer_operation == expected_operation
            assert transfer_job == expected_job

    def test_extract_parts_with_none_operation_name(self):
        """Test extract_parts with None operation name."""
        transfer_operation, transfer_job = CloudStorageTransferLinkHelper.extract_parts(None)

        assert transfer_operation == ""
        assert transfer_job == ""

    def test_extract_parts_with_empty_operation_name(self):
        """Test extract_parts with empty operation name."""
        transfer_operation, transfer_job = CloudStorageTransferLinkHelper.extract_parts("")

        assert transfer_operation == ""
        assert transfer_job == ""

    def test_extract_parts_with_malformed_operation_names(self):
        """Test extract_parts with malformed operation names that might occur in real scenarios."""
        test_cases = [
            ("invalid-format", IndexError),
            ("transferOperations/", IndexError),
            ("transferOperations/123", IndexError),
            ("transferOperations/123-", ("123-", "")),
            ("-transferJobs-job456", IndexError),
            ("transferOperations/123-transferJobs", ("123-transferJobs", "transferJobs")),
        ]

        for operation_name, expected in test_cases:
            if expected is IndexError:
                with pytest.raises(IndexError):
                    CloudStorageTransferLinkHelper.extract_parts(operation_name)
            else:
                transfer_operation, transfer_job = CloudStorageTransferLinkHelper.extract_parts(
                    operation_name
                )
                assert transfer_operation == expected[0]
                assert transfer_job == expected[1]


class TestCloudStorageTransferListLink:
    """Test the CloudStorageTransferListLink with real scenarios."""

    def test_link_properties(self):
        """Test that link properties are set correctly for real usage."""
        link = CloudStorageTransferListLink()

        assert link.name == "Cloud Storage Transfer"
        assert link.key == "cloud_storage_transfer"
        assert "{project_id}" in link.format_str

    def test_format_str_with_real_project_id(self):
        """Test format_str generates correct URL for real project ID."""
        link = CloudStorageTransferListLink()
        formatted_url = link.format_str.format(project_id=REAL_PROJECT_ID)

        assert formatted_url == EXPECTED_LIST_URL
        assert formatted_url.startswith("https://console.cloud.google.com/transfer/jobs?project=")
        assert REAL_PROJECT_ID in formatted_url

    def test_format_str_with_various_project_ids(self):
        """Test format_str with various real project ID formats."""
        project_ids = [
            "my-project",
            "project-123456",
            "my-gcp-project-123",
            "a" * 30,
        ]

        link = CloudStorageTransferListLink()
        for project_id in project_ids:
            formatted_url = link.format_str.format(project_id=project_id)
            assert project_id in formatted_url
            assert formatted_url.startswith("https://console.cloud.google.com/transfer/jobs?")


class TestCloudStorageTransferJobLink:
    """Test the CloudStorageTransferJobLink with real scenarios."""

    def test_link_properties(self):
        """Test that link properties are set correctly for real usage."""
        link = CloudStorageTransferJobLink()

        assert link.name == "Cloud Storage Transfer Job"
        assert link.key == "cloud_storage_transfer_job"
        assert "{project_id}" in link.format_str
        assert "{transfer_job}" in link.format_str

    def test_format_str_with_real_parameters(self):
        """Test format_str generates correct URL for real transfer job."""
        link = CloudStorageTransferJobLink()
        formatted_url = link.format_str.format(project_id=REAL_PROJECT_ID, transfer_job=REAL_TRANSFER_JOB)

        assert formatted_url == EXPECTED_JOB_URL
        assert formatted_url.startswith("https://console.cloud.google.com/transfer/jobs/transferJobs%2F")
        assert REAL_PROJECT_ID in formatted_url
        assert REAL_TRANSFER_JOB in formatted_url

    def test_format_str_url_encoding(self):
        """Test that transfer job ID is properly URL encoded."""
        link = CloudStorageTransferJobLink()
        formatted_url = link.format_str.format(project_id=REAL_PROJECT_ID, transfer_job=REAL_TRANSFER_JOB)

        assert "transferJobs%2F" in formatted_url
        assert "transferJobs/" not in formatted_url


class TestCloudStorageTransferDetailsLink:
    """Test the CloudStorageTransferDetailsLink with real scenarios."""

    def test_link_properties(self):
        """Test that link properties are set correctly for real usage."""
        link = CloudStorageTransferDetailsLink()

        assert link.name == "Cloud Storage Transfer Details"
        assert link.key == "cloud_storage_transfer_details"
        assert "{project_id}" in link.format_str
        assert "{transfer_job}" in link.format_str
        assert "{transfer_operation}" in link.format_str

    def test_format_str_with_real_parameters(self):
        """Test format_str generates correct URL for real transfer operation."""
        link = CloudStorageTransferDetailsLink()

        formatted_url = link.format_str.format(
            project_id=REAL_PROJECT_ID,
            transfer_job="transferJobs",
            transfer_operation="9876543210987654321-transferJobs-1234567890123456789",
        )

        assert formatted_url == EXPECTED_DETAILS_URL
        assert formatted_url.startswith("https://console.cloud.google.com/transfer/jobs/transferJobs%2F")
        assert REAL_PROJECT_ID in formatted_url
        assert "transferJobs" in formatted_url
        assert "9876543210987654321-transferJobs-1234567890123456789" in formatted_url

    def test_extract_parts_with_real_operation_name(self):
        """Test extract_parts with real operation name."""
        transfer_operation, transfer_job = CloudStorageTransferDetailsLink.extract_parts(REAL_OPERATION_NAME)

        assert transfer_operation == "9876543210987654321-transferJobs-1234567890123456789"
        assert transfer_job == "transferJobs"

    def test_extract_parts_edge_cases(self):
        """Test extract_parts with edge cases that might occur in production."""
        edge_cases = [
            ("transferOperations/1-transferJobs-2", "1-transferJobs-2", "transferJobs"),
            ("transferOperations/abc-transferJobs-def", "abc-transferJobs-def", "transferJobs"),
            (
                "transferOperations/123_456-transferJobs-789_012",
                "123_456-transferJobs-789_012",
                "transferJobs",
            ),
        ]

        for operation_name, expected_operation, expected_job in edge_cases:
            transfer_operation, transfer_job = CloudStorageTransferDetailsLink.extract_parts(operation_name)
            assert transfer_operation == expected_operation
            assert transfer_job == expected_job

    def test_persist_method_with_real_operation_name(self):
        """Test persist method behavior with real operation name (unit test)."""
        link = CloudStorageTransferDetailsLink()

        transfer_operation, transfer_job = CloudStorageTransferDetailsLink.extract_parts(REAL_OPERATION_NAME)

        assert transfer_operation == "9876543210987654321-transferJobs-1234567890123456789"
        assert transfer_job == "transferJobs"

        formatted_url = link.format_str.format(
            project_id=REAL_PROJECT_ID,
            transfer_job=transfer_job,
            transfer_operation=transfer_operation,
        )

        assert formatted_url == EXPECTED_DETAILS_URL

    @pytest.mark.db_test
    def test_persist_with_real_operation_name(
        self, create_task_instance_of_operator, session, mock_supervisor_comms
    ):
        """Test persist method with real operation name."""
        from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

        class TestOperator(GoogleCloudBaseOperator):
            operator_extra_links = (CloudStorageTransferDetailsLink(),)

            def __init__(self, project_id: str, **kwargs):
                super().__init__(**kwargs)
                self.project_id = project_id

            @property
            def extra_links_params(self):
                return {
                    "project_id": self.project_id,
                }

            def execute(self, context):
                pass

        ti = create_task_instance_of_operator(
            TestOperator,
            dag_id="test_transfer_dag",
            task_id="test_transfer_task",
            project_id=REAL_PROJECT_ID,
        )
        session.add(ti)
        session.commit()

        from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            from airflow.sdk.execution_time.comms import XComResult

            transfer_operation, transfer_job = CloudStorageTransferDetailsLink.extract_parts(
                REAL_OPERATION_NAME
            )
            mock_supervisor_comms.send.return_value = XComResult(
                key="cloud_storage_transfer_details",
                value={
                    "project_id": REAL_PROJECT_ID,
                    "transfer_job": transfer_job,
                    "transfer_operation": transfer_operation,
                },
            )

        CloudStorageTransferDetailsLink.persist(
            context={"ti": ti, "task": ti.task},
            project_id=REAL_PROJECT_ID,
            operation_name=REAL_OPERATION_NAME,
        )

        actual_url = CloudStorageTransferDetailsLink.get_link(
            CloudStorageTransferDetailsLink(), operator=ti.task, ti_key=ti.key
        )
        assert actual_url == EXPECTED_DETAILS_URL

    @pytest.mark.db_test
    def test_persist_with_none_operation_name(
        self, create_task_instance_of_operator, session, mock_supervisor_comms
    ):
        """Test persist method with None operation name."""
        from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

        class TestOperator(GoogleCloudBaseOperator):
            operator_extra_links = (CloudStorageTransferDetailsLink(),)

            def __init__(self, project_id: str, **kwargs):
                super().__init__(**kwargs)
                self.project_id = project_id

            @property
            def extra_links_params(self):
                return {
                    "project_id": self.project_id,
                }

            def execute(self, context):
                pass

        ti = create_task_instance_of_operator(
            TestOperator,
            dag_id="test_transfer_dag",
            task_id="test_transfer_task",
            project_id=REAL_PROJECT_ID,
        )
        session.add(ti)
        session.commit()

        from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            from airflow.sdk.execution_time.comms import XComResult

            transfer_operation, transfer_job = CloudStorageTransferDetailsLink.extract_parts(None)
            mock_supervisor_comms.send.return_value = XComResult(
                key="cloud_storage_transfer_details",
                value={
                    "project_id": REAL_PROJECT_ID,
                    "transfer_job": transfer_job,
                    "transfer_operation": transfer_operation,
                },
            )

        CloudStorageTransferDetailsLink.persist(
            context={"ti": ti, "task": ti.task},
            project_id=REAL_PROJECT_ID,
            operation_name=None,
        )

        actual_url = CloudStorageTransferDetailsLink.get_link(
            CloudStorageTransferDetailsLink(), operator=ti.task, ti_key=ti.key
        )
        expected_url = f"https://console.cloud.google.com/transfer/jobs/transferJobs%2F/runs/transferOperations%2F?project={REAL_PROJECT_ID}"
        assert actual_url == expected_url

    @pytest.mark.db_test
    def test_persist_with_empty_operation_name(
        self, create_task_instance_of_operator, session, mock_supervisor_comms
    ):
        """Test persist method with empty operation name."""
        from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator

        class TestOperator(GoogleCloudBaseOperator):
            operator_extra_links = (CloudStorageTransferDetailsLink(),)

            def __init__(self, project_id: str, **kwargs):
                super().__init__(**kwargs)
                self.project_id = project_id

            @property
            def extra_links_params(self):
                return {
                    "project_id": self.project_id,
                }

            def execute(self, context):
                pass

        ti = create_task_instance_of_operator(
            TestOperator,
            dag_id="test_transfer_dag",
            task_id="test_transfer_task",
            project_id=REAL_PROJECT_ID,
        )
        session.add(ti)
        session.commit()

        from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

        if AIRFLOW_V_3_0_PLUS and mock_supervisor_comms:
            from airflow.sdk.execution_time.comms import XComResult

            transfer_operation, transfer_job = CloudStorageTransferDetailsLink.extract_parts("")
            mock_supervisor_comms.send.return_value = XComResult(
                key="cloud_storage_transfer_details",
                value={
                    "project_id": REAL_PROJECT_ID,
                    "transfer_job": transfer_job,
                    "transfer_operation": transfer_operation,
                },
            )

        CloudStorageTransferDetailsLink.persist(
            context={"ti": ti, "task": ti.task},
            project_id=REAL_PROJECT_ID,
            operation_name="",
        )

        actual_url = CloudStorageTransferDetailsLink.get_link(
            CloudStorageTransferDetailsLink(), operator=ti.task, ti_key=ti.key
        )
        expected_url = f"https://console.cloud.google.com/transfer/jobs/transferJobs%2F/runs/transferOperations%2F?project={REAL_PROJECT_ID}"
        assert actual_url == expected_url


class TestIntegrationScenarios:
    """Integration tests with real-world scenarios."""

    def test_complete_transfer_workflow_links(self):
        """Test all link types work together in a complete transfer workflow."""
        list_link = CloudStorageTransferListLink()
        list_url = list_link.format_str.format(project_id=REAL_PROJECT_ID)

        job_link = CloudStorageTransferJobLink()
        job_url = job_link.format_str.format(project_id=REAL_PROJECT_ID, transfer_job=REAL_TRANSFER_JOB)

        details_link = CloudStorageTransferDetailsLink()
        details_url = details_link.format_str.format(
            project_id=REAL_PROJECT_ID,
            transfer_job="transferJobs",
            transfer_operation="9876543210987654321-transferJobs-1234567890123456789",
        )

        assert all(
            url.startswith("https://console.cloud.google.com/transfer")
            for url in [list_url, job_url, details_url]
        )
        assert REAL_PROJECT_ID in list_url
        assert REAL_PROJECT_ID in job_url
        assert REAL_PROJECT_ID in details_url
        assert REAL_TRANSFER_JOB in job_url
        assert "transferJobs" in details_url
        assert "9876543210987654321-transferJobs-1234567890123456789" in details_url

    def test_url_consistency_across_link_types(self):
        """Test that URLs are consistent across different link types."""
        base_url = "https://console.cloud.google.com/transfer"

        list_link = CloudStorageTransferListLink()
        job_link = CloudStorageTransferJobLink()
        details_link = CloudStorageTransferDetailsLink()

        assert list_link.format_str.startswith(base_url)
        assert job_link.format_str.startswith(base_url)
        assert details_link.format_str.startswith(base_url)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging

from google.cloud.bigquery import CopyJob, ExtractJob, LoadJob, QueryJob

from airflow.providers.common.compat.lineage.hook import get_hook_lineage_collector
from airflow.providers.common.sql.hooks.lineage import send_sql_hook_lineage

log = logging.getLogger(__name__)


def _add_bq_table_to_lineage(collector, context, table_ref, *, is_input: bool):
    method = collector.add_input_asset if is_input else collector.add_output_asset
    method(
        context=context,
        scheme="bigquery",
        asset_kwargs={
            "project_id": table_ref.project,
            "dataset_id": table_ref.dataset_id,
            "table_id": table_ref.table_id,
        },
    )


def _add_gcs_uris_to_lineage(collector, context, uris, *, is_input: bool):
    method = collector.add_input_asset if is_input else collector.add_output_asset
    for uri in uris or []:
        method(context=context, uri=uri)


def send_hook_lineage_for_bq_job(context, job):
    """
    Send hook-level lineage for a BigQuery job to the lineage collector.

    Handles all four BigQuery job types:
    - QUERY: delegates to send_sql_hook_lineage for SQL parsing
    - LOAD: source URIs (GCS) as inputs, destination table as output
    - COPY: source tables as inputs, destination table as output
    - EXTRACT: source table as input, destination URIs (GCS) as outputs

    :param context: The hook instance used as lineage context.
    :param job: A BigQuery job object (QueryJob, LoadJob, CopyJob, or ExtractJob).
    """
    collector = get_hook_lineage_collector()

    if isinstance(job, QueryJob):
        log.debug("Sending Hook Level Lineage for Query job.")
        send_sql_hook_lineage(
            context=context,
            sql=job.query,
            job_id=job.job_id,
            default_db=job.default_dataset.project if job.default_dataset else None,
            default_schema=job.default_dataset.dataset_id if job.default_dataset else None,
        )
        return

    try:
        if isinstance(job, LoadJob):
            log.debug("Sending Hook Level Lineage for Load job.")
            _add_gcs_uris_to_lineage(collector, context, job.source_uris, is_input=True)
            if job.destination:
                _add_bq_table_to_lineage(collector, context, job.destination, is_input=False)
        elif isinstance(job, CopyJob):
            log.debug("Sending Hook Level Lineage for Copy job.")
            for source_table in job.sources or []:
                _add_bq_table_to_lineage(collector, context, source_table, is_input=True)
            if job.destination:
                _add_bq_table_to_lineage(collector, context, job.destination, is_input=False)
        elif isinstance(job, ExtractJob):
            log.debug("Sending Hook Level Lineage for Extract job.")
            if job.source:
                _add_bq_table_to_lineage(collector, context, job.source, is_input=True)
            _add_gcs_uris_to_lineage(collector, context, job.destination_uris, is_input=False)
    except Exception as e:
        log.warning("Sending BQ job hook level lineage failed: %s", f"{e.__class__.__name__}: {str(e)}")
        log.debug("Exception details:", exc_info=True)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from enum import Enum


class DataFusionPipelineType(Enum):
    """Enum for Data Fusion pipeline types."""

    BATCH = "batch"
    STREAM = "stream"

    @staticmethod
    def from_str(value: str) -> DataFusionPipelineType:
        value_to_item = {item.value: item for item in DataFusionPipelineType}
        if value in value_to_item:
            return value_to_item[value]
        raise ValueError(f"Invalid value '{value}'. Valid values are: {[i for i in value_to_item.keys()]}")
